"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from decimal import Decimal
from datetime import timedelta
from selenium.webdriver.support.ui import Select
from selenium.webdriver.common.by import By

from biweeklybudget.utils import dtnow
from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.models.fuel import Vehicle, FuelFill
from biweeklybudget.models.transaction import Transaction


LEVEL_OPTS = [
    ['0', '0/10'], ['10', '1/10'], ['20', '2/10'], ['30', '3/10'],
    ['40', '4/10'], ['50', '5/10'], ['60', '6/10'], ['70', '7/10'],
    ['80', '8/10'], ['90', '9/10'], ['100', '10/10']
]


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestFuel(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/fuel')

    def test_heading(self, selenium):
        heading = selenium.find_element(By.CLASS_NAME, 'navbar-brand')
        assert heading.text == 'Fuel Log - BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element(By.ID, 'side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element(By.ID, 'notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
@pytest.mark.incremental
class TestFuelLogView(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.dt = dtnow()
        self.get(selenium, base_url + '/fuel')

    def test_01_table(self, selenium):
        table = selenium.find_element(By.ID, 'table-fuel-log')
        texts = self.tbody2textlist(table)
        assert texts == [
            [
                (self.dt + timedelta(days=1)).date().strftime('%Y-%m-%d'),
                'Veh1 (1)',
                '1,011',
                '111',
                '10',
                '11%',
                '89%',
                'fill_loc v1 1',
                '$2.11',
                '$4.22',
                '11.000',
                '21.1',
                '0.909',
                'notes v1 1'
            ],
            [
                (self.dt + timedelta(days=1)).date().strftime('%Y-%m-%d'),
                'Veh2 (2)',
                '1,012',
                '112',
                '10',
                '12%',
                '88%',
                'fill_loc v2 1',
                '$2.12',
                '$4.24',
                '12.000',
                '21.2',
                '0.833',
                'notes v2 1'
            ],
            [
                self.dt.date().strftime('%Y-%m-%d'),
                'Veh1 (1)',
                '1,001',
                '101',
                '',
                '1%',
                '99%',
                'fill_loc v1 0',
                '$2.01',
                '$2.01',
                '1.000',
                '20.1',
                '',
                'notes v1 0'
            ],
            [
                self.dt.date().strftime('%Y-%m-%d'),
                'Veh2 (2)',
                '1,002',
                '102',
                '',
                '2%',
                '98%',
                'fill_loc v2 0',
                '$2.02',
                '$2.02',
                '2.000',
                '20.2',
                '',
                'notes v2 0'
            ]
        ]

    def test_02_filter_opts(self, selenium):
        self.get(selenium, self.baseurl + '/fuel')
        veh_filter = Select(selenium.find_element(By.ID, 'vehicle_filter'))
        # find the options
        opts = []
        for o in veh_filter.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'Veh1'],
            ['2', 'Veh2']
        ]

    def test_03_filter(self, selenium):
        p1odo = [
            '1,011',
            '1,012',
            '1,001',
            '1,002'
        ]
        self.get(selenium, self.baseurl + '/fuel')
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-fuel-log')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        odo = [t[2] for t in texts]
        # check sanity
        assert odo == p1odo
        veh_filter = Select(selenium.find_element(By.ID, 'vehicle_filter'))
        # select Veh1
        veh_filter.select_by_value('1')
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-fuel-log')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        odo = [t[2] for t in texts]
        assert odo == ['1,011', '1,001']
        # select back to all
        veh_filter.select_by_value('None')
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-fuel-log')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        odo = [t[2] for t in texts]
        # check sanity
        assert odo == p1odo

    def test_04_search(self, selenium):
        self.get(selenium, self.baseurl + '/fuel')
        search = self.retry_stale(
            lambda: selenium.find_element(By.XPATH, '//input[@type="search"]')
        )
        search.send_keys('v1 1')
        self.wait_for_jquery_done(selenium)
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-fuel-log')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        odo = [t[2] for t in texts]
        assert odo == ['1,011']

    def test_05_vehicles(self, base_url, selenium):
        self.get(selenium, self.baseurl + '/fuel')
        table = selenium.find_element(By.ID, 'vehicles-table')
        elems = self.tbody2elemlist(table)
        htmls = []
        for row in elems:
            htmls.append(
                [x.get_attribute('innerHTML') for x in row]
            )
        assert htmls == [
            [
                '<a href="javascript:vehicleModal(1)">1</a>',
                '<a href="javascript:vehicleModal(1)">Veh1</a>',
                'True'
            ],
            [
                '<a href="javascript:vehicleModal(2)">2</a>',
                '<a href="javascript:vehicleModal(2)">Veh2</a>',
                'True'
            ],
            [
                '<a href="javascript:vehicleModal(3)">3</a>',
                '<a href="javascript:vehicleModal(3)">Veh3Inactive</a>',
                'False'
            ]
        ]


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestModals(AcceptanceHelper):

    def test_00_vehicle_verify_db(self, testdb):
        b = testdb.query(Vehicle).get(1)
        assert b is not None
        assert b.name == 'Veh1'
        assert b.is_active is True
        b = testdb.query(Vehicle).get(3)
        assert b is not None
        assert b.name == 'Veh3Inactive'
        assert b.is_active is False

    def test_01_vehicle_populate_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/fuel')
        link = selenium.find_element(By.XPATH, '//a[text()="Veh1"]')
        self.wait_until_clickable(selenium, '//a[text()="Veh1"]', by='xpath')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Vehicle 1'
        assert selenium.find_element(By.ID, 'vehicle_frm_id').get_attribute(
            'value') == '1'
        assert selenium.find_element(By.ID,
                                     'vehicle_frm_name').get_attribute('value') == 'Veh1'
        assert selenium.find_element(By.ID, 'vehicle_frm_active').is_selected()

    def test_02_vehicle_edit_modal_inactive(self, base_url, selenium):
        self.get(selenium, base_url + '/fuel')
        link = selenium.find_element(By.XPATH, '//a[text()="Veh3Inactive"]')
        self.wait_until_clickable(
            selenium, '//a[text()="Veh3Inactive"]', by='xpath'
        )
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Vehicle 3'
        assert selenium.find_element(By.ID, 'vehicle_frm_id').get_attribute(
            'value') == '3'
        assert selenium.find_element(By.ID,
                                     'vehicle_frm_name').get_attribute('value') == 'Veh3Inactive'
        assert selenium.find_element(By.ID,
                                     'vehicle_frm_active').is_selected() is False
        name = selenium.find_element(By.ID, 'vehicle_frm_name')
        name.clear()
        name.send_keys('Veh3Edited')
        selenium.find_element(By.ID, 'vehicle_frm_active').click()
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Vehicle 3 in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        self.wait_for_id(selenium, 'vehicles-table')
        # test that the table is updated
        table = selenium.find_element(By.ID, 'vehicles-table')
        elems = self.tbody2elemlist(table)
        htmls = []
        for row in elems:
            htmls.append(
                [y.get_attribute('innerHTML') for y in row]
            )
        assert htmls == [
            [
                '<a href="javascript:vehicleModal(1)">1</a>',
                '<a href="javascript:vehicleModal(1)">Veh1</a>',
                'True'
            ],
            [
                '<a href="javascript:vehicleModal(2)">2</a>',
                '<a href="javascript:vehicleModal(2)">Veh2</a>',
                'True'
            ],
            [
                '<a href="javascript:vehicleModal(3)">3</a>',
                '<a href="javascript:vehicleModal(3)">Veh3Edited</a>',
                'True'
            ]
        ]

    def test_03_vehicle_verify_db(self, testdb):
        b = testdb.query(Vehicle).get(3)
        assert b is not None
        assert b.name == 'Veh3Edited'
        assert b.is_active is True

    def test_04_vehicle_modal_add(self, base_url, selenium):
        self.get(selenium, base_url + '/fuel')
        link = selenium.find_element(By.ID, 'btn-add-vehicle')
        self.wait_until_clickable(selenium, 'btn-add-vehicle')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Vehicle'
        name = selenium.find_element(By.ID, 'vehicle_frm_name')
        name.clear()
        name.send_keys('Vehicle4')
        selenium.find_element(By.ID, 'vehicle_frm_active').click()
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Vehicle 4 in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        self.wait_for_id(selenium, 'vehicles-table')
        # test that the table is updated
        table = selenium.find_element(By.ID, 'vehicles-table')
        elems = self.tbody2elemlist(table)
        htmls = []
        for row in elems:
            htmls.append(
                [y.get_attribute('innerHTML') for y in row]
            )
        assert htmls == [
            [
                '<a href="javascript:vehicleModal(1)">1</a>',
                '<a href="javascript:vehicleModal(1)">Veh1</a>',
                'True'
            ],
            [
                '<a href="javascript:vehicleModal(2)">2</a>',
                '<a href="javascript:vehicleModal(2)">Veh2</a>',
                'True'
            ],
            [
                '<a href="javascript:vehicleModal(3)">3</a>',
                '<a href="javascript:vehicleModal(3)">Veh3Edited</a>',
                'True'
            ],
            [
                '<a href="javascript:vehicleModal(4)">4</a>',
                '<a href="javascript:vehicleModal(4)">Vehicle4</a>',
                'False'
            ]
        ]

    def test_10_fuel_verify_db(self, testdb):
        ids = [
            t.id for t in testdb.query(FuelFill).all()
        ]
        assert len(ids) == 6
        assert max(ids) == 6
        trans_ids = [
            x.id for x in testdb.query(Transaction).all()
        ]
        assert len(trans_ids) == 4
        assert max(trans_ids) == 4

    def test_11_fuel_populate_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/fuel')
        link = selenium.find_element(By.ID, 'btn-add-fuel')
        self.wait_until_clickable(selenium, 'btn-add-fuel')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add Fuel Fill'
        veh_sel = Select(selenium.find_element(By.ID, 'fuel_frm_vehicle'))
        opts = [[o.get_attribute('value'), o.text] for o in veh_sel.options]
        assert opts == [['1', 'Veh1'], ['2', 'Veh2'], ['3', 'Veh3Edited']]
        assert veh_sel.first_selected_option.get_attribute('value') == '1'
        date = selenium.find_element(By.ID, 'fuel_frm_date')
        assert date.get_attribute(
            'value') == dtnow().date().strftime('%Y-%m-%d')
        odo = selenium.find_element(By.ID, 'fuel_frm_odo_miles')
        assert odo.get_attribute('value') == ''
        rep_mi = selenium.find_element(By.ID, 'fuel_frm_reported_miles')
        assert rep_mi.get_attribute('value') == ''
        lvl_before = Select(
            selenium.find_element(By.ID, 'fuel_frm_level_before')
        )
        opts = [[o.get_attribute('value'), o.text] for o in lvl_before.options]
        assert opts == LEVEL_OPTS
        assert lvl_before.first_selected_option.get_attribute('value') == '0'
        lvl_after = Select(
            selenium.find_element(By.ID, 'fuel_frm_level_after')
        )
        opts = [[o.get_attribute('value'), o.text] for o in lvl_after.options]
        assert opts == LEVEL_OPTS
        assert lvl_after.first_selected_option.get_attribute('value') == '100'
        fill_loc = selenium.find_element(By.ID, 'fuel_frm_fill_loc')
        assert fill_loc.get_attribute('value') == ''
        cpg = selenium.find_element(By.ID, 'fuel_frm_cost_per_gallon')
        assert cpg.get_attribute('value') == ''
        cost = selenium.find_element(By.ID, 'fuel_frm_total_cost')
        assert cost.get_attribute('value') == ''
        gals = selenium.find_element(By.ID, 'fuel_frm_gallons')
        assert gals.get_attribute('value') == ''
        rep_mpg = selenium.find_element(By.ID, 'fuel_frm_reported_mpg')
        assert rep_mpg.get_attribute('value') == ''
        add_trans = selenium.find_element(By.ID, 'fuel_frm_add_trans')
        assert add_trans.is_selected()
        acct_sel = Select(body.find_element(By.ID, 'fuel_frm_account'))
        opts = [[o.get_attribute('value'), o.text] for o in acct_sel.options]
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        budget_sel = Select(body.find_element(By.ID, 'fuel_frm_budget'))
        opts = [[o.get_attribute('value'), o.text] for o in budget_sel.options]
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['3', 'Periodic3 Inactive'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['6', 'Standing3 Inactive']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '2'
        notes = selenium.find_element(By.ID, 'fuel_frm_notes')
        assert notes.get_attribute('value') == ''

    def test_12_fuel_add_no_trans(self, base_url, selenium):
        self.get(selenium, base_url + '/fuel')
        link = selenium.find_element(By.ID, 'btn-add-fuel')
        self.wait_until_clickable(selenium, 'btn-add-fuel')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add Fuel Fill'
        veh_sel = Select(selenium.find_element(By.ID, 'fuel_frm_vehicle'))
        veh_sel.select_by_value('2')
        date = selenium.find_element(By.ID, 'fuel_frm_date')
        date.clear()
        date.send_keys(
            (dtnow() - timedelta(days=3)).date().strftime('%Y-%m-%d')
        )
        odo = selenium.find_element(By.ID, 'fuel_frm_odo_miles')
        odo.clear()
        odo.send_keys('1123')
        rep_mi = selenium.find_element(By.ID, 'fuel_frm_reported_miles')
        rep_mi.clear()
        rep_mi.send_keys('123')
        lvl_before = Select(
            selenium.find_element(By.ID, 'fuel_frm_level_before')
        )
        lvl_before.select_by_value('50')
        lvl_after = Select(
            selenium.find_element(By.ID, 'fuel_frm_level_after')
        )
        lvl_after.select_by_value('90')
        fill_loc = selenium.find_element(By.ID, 'fuel_frm_fill_loc')
        fill_loc.clear()
        fill_loc.send_keys('Fill Location')
        cpg = selenium.find_element(By.ID, 'fuel_frm_cost_per_gallon')
        cpg.clear()
        cpg.send_keys('1.239')
        cost = selenium.find_element(By.ID, 'fuel_frm_total_cost')
        cost.clear()
        cost.send_keys('12.34')
        gals = selenium.find_element(By.ID, 'fuel_frm_gallons')
        gals.clear()
        gals.send_keys('6.789')
        rep_mpg = selenium.find_element(By.ID, 'fuel_frm_reported_mpg')
        rep_mpg.clear()
        rep_mpg.send_keys('34.5')
        add_trans = selenium.find_element(By.ID, 'fuel_frm_add_trans')
        add_trans.click()
        assert add_trans.is_selected() is False
        acct_sel = Select(body.find_element(By.ID, 'fuel_frm_account'))
        acct_sel.select_by_value('3')
        budget_sel = Select(body.find_element(By.ID, 'fuel_frm_budget'))
        budget_sel.select_by_value('1')
        notes = selenium.find_element(By.ID, 'fuel_frm_notes')
        notes.clear()
        notes.send_keys('My Notes')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved FuelFill 7 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that datatable was updated
        # Wait for DataTable to fully render (may take time in Docker due to
        # network latency between test and container)
        expected_odo = [
            '1,011', '1,012', '1,013', '1,001', '1,002', '1,003', '1,123'
        ]
        for _ in range(10):  # Retry up to 10 times
            table = selenium.find_element(By.ID, 'table-fuel-log')
            odo_reads = [y[2] for y in self.tbody2textlist(table)]
            # Check if all cells are populated (no empty strings)
            if '' not in odo_reads and odo_reads == expected_odo:
                break
            from time import sleep
            sleep(0.5)
        assert odo_reads == expected_odo
        notif = selenium.find_element(By.ID, 'last_mpg_notice')
        assert notif.get_attribute('innerHTML') == 'Last fill MPG: 16.349'

    def test_13_fuel_verify_db(self, testdb):
        ids = [
            t.id for t in testdb.query(FuelFill).all()
        ]
        assert len(ids) == 7
        assert max(ids) == 7
        fill = testdb.query(FuelFill).get(7)
        assert fill.date == (dtnow() - timedelta(days=3)).date()
        assert fill.vehicle_id == 2
        assert fill.odometer_miles == 1123
        assert fill.reported_miles == 123
        assert fill.level_before == 50
        assert fill.level_after == 90
        assert fill.fill_location == 'Fill Location'
        assert fill.cost_per_gallon == Decimal('1.239')
        assert fill.total_cost == Decimal('12.34')
        assert fill.gallons == Decimal('6.789')
        assert fill.reported_mpg == Decimal('34.5')
        assert fill.notes == 'My Notes'
        # calculated values
        assert fill.calculated_miles == 111
        assert fill.calculated_mpg == Decimal('16.349')
        trans_ids = [
            x.id for x in testdb.query(Transaction).all()
        ]
        assert len(trans_ids) == 4
        assert max(trans_ids) == 4

    def test_14_fuel_add_with_trans(self, base_url, selenium):
        self.get(selenium, base_url + '/fuel')
        link = selenium.find_element(By.ID, 'btn-add-fuel')
        self.wait_until_clickable(selenium, 'btn-add-fuel')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add Fuel Fill'
        veh_sel = Select(selenium.find_element(By.ID, 'fuel_frm_vehicle'))
        veh_sel.select_by_value('1')
        date = selenium.find_element(By.ID, 'fuel_frm_date')
        date.clear()
        date.send_keys(
            (dtnow() - timedelta(days=2)).date().strftime('%Y-%m-%d')
        )
        odo = selenium.find_element(By.ID, 'fuel_frm_odo_miles')
        odo.clear()
        odo.send_keys('1256')
        rep_mi = selenium.find_element(By.ID, 'fuel_frm_reported_miles')
        rep_mi.clear()
        rep_mi.send_keys('345')
        lvl_before = Select(
            selenium.find_element(By.ID, 'fuel_frm_level_before')
        )
        lvl_before.select_by_value('10')
        lvl_after = Select(
            selenium.find_element(By.ID, 'fuel_frm_level_after')
        )
        lvl_after.select_by_value('100')
        fill_loc = selenium.find_element(By.ID, 'fuel_frm_fill_loc')
        fill_loc.clear()
        fill_loc.send_keys('Fill Location2')
        cpg = selenium.find_element(By.ID, 'fuel_frm_cost_per_gallon')
        cpg.clear()
        cpg.send_keys('1.459')
        cost = selenium.find_element(By.ID, 'fuel_frm_total_cost')
        cost.clear()
        cost.send_keys('14.82')
        gals = selenium.find_element(By.ID, 'fuel_frm_gallons')
        gals.clear()
        gals.send_keys('5.678')
        rep_mpg = selenium.find_element(By.ID, 'fuel_frm_reported_mpg')
        rep_mpg.clear()
        rep_mpg.send_keys('28.3')
        add_trans = selenium.find_element(By.ID, 'fuel_frm_add_trans')
        assert add_trans.is_selected()
        acct_sel = Select(body.find_element(By.ID, 'fuel_frm_account'))
        acct_sel.select_by_value('3')
        budget_sel = Select(body.find_element(By.ID, 'fuel_frm_budget'))
        budget_sel.select_by_value('1')
        notes = selenium.find_element(By.ID, 'fuel_frm_notes')
        notes.clear()
        notes.send_keys('My Notes2')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved FuelFill 8 ' \
                                 'and Transaction 5 in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that datatable was updated
        # Wait for DataTable to fully render (may take time in Docker due to
        # network latency between test and container)
        expected_odo = [
            '1,011', '1,012', '1,013',
            '1,001', '1,002', '1,003',
            '1,256', '1,123'
        ]
        for _ in range(10):  # Retry up to 10 times
            table = selenium.find_element(By.ID, 'table-fuel-log')
            odo_reads = [y[2] for y in self.tbody2textlist(table)]
            # Check if all cells are populated (no empty strings)
            if '' not in odo_reads and odo_reads == expected_odo:
                break
            from time import sleep
            sleep(0.5)
        assert odo_reads == expected_odo

    def test_15_fuel_verify_db(self, testdb):
        ids = [
            t.id for t in testdb.query(FuelFill).all()
        ]
        assert len(ids) == 8
        assert max(ids) == 8
        fill = testdb.query(FuelFill).get(8)
        assert fill.date == (dtnow() - timedelta(days=2)).date()
        assert fill.vehicle_id == 1
        assert fill.odometer_miles == 1256
        assert fill.reported_miles == 345
        assert fill.level_before == 10
        assert fill.level_after == 100
        assert fill.fill_location == 'Fill Location2'
        assert fill.cost_per_gallon == Decimal('1.459')
        assert fill.total_cost == Decimal('14.82')
        assert fill.gallons == Decimal('5.678')
        assert fill.reported_mpg == Decimal('28.3')
        assert fill.notes == 'My Notes2'
        # calculated values
        assert fill.calculated_miles == 245
        assert fill.calculated_mpg == Decimal('43.148')
        trans_ids = [
            x.id for x in testdb.query(Transaction).all()
        ]
        assert len(trans_ids) == 5
        assert max(trans_ids) == 5
        trans = testdb.query(Transaction).get(5)
        assert trans.date == (dtnow() - timedelta(days=2)).date()
        assert trans.actual_amount == Decimal('14.82')
        assert trans.budgeted_amount is None
        assert trans.description == 'Fill Location2 - FuelFill #8 (Veh1)'
        assert trans.notes == 'My Notes2'
        assert trans.account_id == 3
        assert trans.planned_budget_id is None
        assert trans.scheduled_trans_id is None
        assert len(trans.budget_transactions) == 1
        assert trans.budget_transactions[0].budget_id == 1
        assert trans.budget_transactions[0].amount == Decimal('14.82')



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging

logger = logging.getLogger(__name__)


def set_browser_for_fullpage_screenshot(driver, extra_height_px=100):
    """
    Helper method to resize a Selenium browser to the size of the document,
    for taking a full page screenshot.

    :param extra_height_px: extra pixels to add to resize height
    :param driver: Selenium webdriver instance
    """
    height = driver.execute_script(
        "return Math.max(document.body.scrollHeight, "
        "document.body.offsetHeight, document.documentElement."
        "clientHeight, document.documentElement.scrollHeight, "
        "document.documentElement.offsetHeight);"
    )
    height += extra_height_px
    curr_size = driver.get_window_size()  # dict, "width" and "height" keys
    logger.debug(
        'Resizing browser height from %d to %d', curr_size['height'], height
    )
    driver.set_window_size(curr_size['width'], height)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from datetime import datetime, timedelta
from pytz import UTC
from decimal import Decimal

from selenium.webdriver.support.ui import Select
from selenium.webdriver.common.by import By
from biweeklybudget.utils import dtnow
from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.tests.sqlhelpers import restore_mysqldump
from biweeklybudget.tests.conftest import get_db_engine
from biweeklybudget.models import *
from biweeklybudget.settings import PAY_PERIOD_START_DATE
from biweeklybudget.biweeklypayperiod import BiweeklyPayPeriod


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestIndexNavigation(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url)

    def test_heading(self, selenium):
        heading = selenium.find_element(By.CLASS_NAME, 'navbar-brand')
        assert heading.text == 'BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element(By.ID, 'side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element(By.ID, 'notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestIndexAccounts(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url)

    def test_bank_table(self, selenium):
        table = selenium.find_element(By.XPATH,
                                      "//div[@id='panel-bank-accounts']//table"
                                      )
        assert self.thead2list(table) == [
            'Account', 'Balance', 'Unreconciled', 'Difference'
        ]
        assert self.tbody2textlist(table) == [
            ['BankOne', '$12,789.01 (14 hours ago)', '$0.00', '$12,789.01'],
            ['BankTwoStale', '$100.23 (a month ago)', '-$333.33', '$433.56']
        ]
        links = []
        tbody = table.find_element(By.TAG_NAME, 'tbody')
        for tr in tbody.find_elements(By.TAG_NAME, 'tr'):
            td = tr.find_elements(By.TAG_NAME, 'td')[0]
            links.append(td.get_attribute('innerHTML'))
        assert links == [
            '<a href="/accounts/1">BankOne</a>',
            '<a href="/accounts/2">BankTwoStale</a>',
        ]

    def test_bank_stale_span(self, selenium):
        tbody = selenium.find_element(By.XPATH,
                                      "//div[@id='panel-bank-accounts']//table/tbody"
                                      )
        rows = tbody.find_elements(By.TAG_NAME, 'tr')
        bankTwoStale_bal_td = rows[1].find_elements(By.TAG_NAME, 'td')[1]
        bal_span = bankTwoStale_bal_td.find_elements(By.TAG_NAME, 'span')[1]
        assert bal_span.text == '(a month ago)'
        assert bal_span.get_attribute('class') == 'data_age text-danger'

    def test_credit_table(self, selenium):
        table = selenium.find_element(By.XPATH,
                                      "//div[@id='panel-credit-cards']//table"
                                      )
        assert self.thead2list(table) == [
            'Account', 'Balance', 'Available', 'Avail - Unrec'
        ]
        assert self.tbody2textlist(table) == [
            ['CreditOne', '-$952.06 (13 hours ago)', '$1,047.94', '$503.40'],
            ['CreditTwo', '-$5,498.65 (a day ago)', '$1.35', '$1.35']
        ]
        links = []
        tbody = table.find_element(By.TAG_NAME, 'tbody')
        for tr in tbody.find_elements(By.TAG_NAME, 'tr'):
            td = tr.find_elements(By.TAG_NAME, 'td')[0]
            links.append(td.get_attribute('innerHTML'))
        assert links == [
            '<a href="/accounts/3">CreditOne</a>',
            '<a href="/accounts/4">CreditTwo</a>',
        ]

    def test_investment_table(self, selenium):
        table = selenium.find_element(By.XPATH,
                                      "//div[@id='panel-investment']//table"
                                      )
        assert self.thead2list(table) == ['Account', 'Value']
        assert self.tbody2textlist(table) == [
            ['InvestmentOne', '$10,362.91 (13 days ago)']
        ]
        links = []
        tbody = table.find_element(By.TAG_NAME, 'tbody')
        for tr in tbody.find_elements(By.TAG_NAME, 'tr'):
            td = tr.find_elements(By.TAG_NAME, 'td')[0]
            links.append(td.get_attribute('innerHTML'))
        assert links == [
            '<a href="/accounts/5">InvestmentOne</a>'
        ]


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestIndexBudgets(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):  # noqa
        self.baseurl = base_url
        self.get(selenium, base_url + '/')

    def test_budgets_table(self, selenium):
        stable = selenium.find_element(By.ID, 'table-standing-budgets')
        stexts = self.tbody2textlist(stable)
        assert stexts == [
            ['Standing1 (4)', '$1,284.23'],
            ['Standing2 (5)', '$9,482.29']
        ]
        selems = self.tbody2elemlist(stable)
        assert selems[0][0].get_attribute(
            'innerHTML') == '<a href="/budgets/4">' \
                            'Standing1 (4)</a>'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
@pytest.mark.incremental
class TestIndexPayPeriods(AcceptanceHelper):

    def test_0_clean_db(self, dump_file_path):
        # clean the database; empty schema
        restore_mysqldump(dump_file_path, get_db_engine(), with_data=False)

    def test_1_add_account(self, testdb):
        a = Account(
            description='First Bank Account',
            name='BankOne',
            ofx_cat_memo_to_name=True,
            ofxgetter_config_json='{"foo": "bar"}',
            vault_creds_path='secret/foo/bar/BankOne',
            acct_type=AcctType.Bank
        )
        testdb.add(a)
        a.set_balance(
            overall_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC),
            ledger=Decimal('1.0'),
            ledger_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC)
        )
        testdb.flush()
        testdb.commit()

    def test_2_add_budgets(self, testdb):
        testdb.add(Budget(
            name='1Income',
            is_periodic=True,
            description='1Income',
            starting_balance=Decimal('1000.00'),
            is_income=True
        ))
        testdb.add(Budget(
            name='2Periodic',
            is_periodic=True,
            description='2Periodic',
            starting_balance=Decimal('500.00')
        ))
        testdb.add(Budget(
            name='3Periodic',
            is_periodic=True,
            description='3Periodic',
            starting_balance=Decimal('0.00')
        ))
        testdb.flush()
        testdb.commit()

    def pay_periods(self, db):
        x = BiweeklyPayPeriod.period_for_date(
            (PAY_PERIOD_START_DATE - timedelta(days=2)), db
        )
        periods = []
        for i in range(0, 10):
            periods.append(x)
            x = x.next
        return periods

    def test_3_add_transactions(self, testdb):
        acct = testdb.query(Account).get(1)
        ibudget = testdb.query(Budget).get(1)
        e1budget = testdb.query(Budget).get(2)
        e2budget = testdb.query(Budget).get(3)
        periods = self.pay_periods(testdb)
        # previous pay period
        ppdate = periods[0].start_date
        t1 = Transaction(
            date=(ppdate + timedelta(days=1)),
            budget_amounts={ibudget: Decimal('100.00')},
            budgeted_amount=Decimal('100.00'),
            description='prev income',
            account=acct,
            planned_budget=ibudget
        )
        testdb.add(t1)
        t2 = Transaction(
            date=(ppdate + timedelta(days=2)),
            budget_amounts={e2budget: Decimal('250.00')},
            description='prev trans 1',
            account=acct
        )
        testdb.add(t2)
        t3 = Transaction(
            date=(ppdate + timedelta(days=3)),
            budget_amounts={e1budget: Decimal('600.00')},
            budgeted_amount=Decimal('500.00'),
            description='prev trans 2',
            account=acct,
            planned_budget=e1budget
        )
        testdb.add(t3)
        ppdate = periods[1].start_date
        t4 = Transaction(
            date=(ppdate + timedelta(days=1)),
            budget_amounts={ibudget: Decimal('1400.00')},
            budgeted_amount=Decimal('100.00'),
            description='prev income',
            account=acct,
            planned_budget=ibudget
        )
        testdb.add(t4)
        t5 = Transaction(
            date=(ppdate + timedelta(days=2)),
            budget_amounts={e2budget: Decimal('1850.00')},
            description='prev trans 1',
            account=acct
        )
        testdb.add(t5)
        t6 = Transaction(
            date=(ppdate + timedelta(days=3)),
            budget_amounts={e1budget: Decimal('600.00')},
            budgeted_amount=Decimal('500.00'),
            description='prev trans 2',
            account=acct,
            planned_budget=e1budget
        )
        testdb.add(t6)
        ppdate = periods[2].start_date
        t7 = Transaction(
            date=(ppdate + timedelta(days=1)),
            budget_amounts={ibudget: Decimal('1400.00')},
            budgeted_amount=Decimal('100.00'),
            description='prev income',
            account=acct,
            planned_budget=ibudget
        )
        testdb.add(t7)
        t8 = Transaction(
            date=(ppdate + timedelta(days=2)),
            budget_amounts={e2budget: Decimal('788.00')},
            description='prev trans 1',
            account=acct
        )
        testdb.add(t8)
        t9 = Transaction(
            date=(ppdate + timedelta(days=3)),
            budget_amounts={e1budget: Decimal('600.00')},
            budgeted_amount=Decimal('500.00'),
            description='prev trans 2',
            account=acct,
            planned_budget=e1budget
        )
        testdb.add(t9)
        ppdate = periods[3].start_date
        t10 = Transaction(
            date=(ppdate + timedelta(days=1)),
            budget_amounts={ibudget: Decimal('1400.00')},
            budgeted_amount=Decimal('100.00'),
            description='prev income',
            account=acct,
            planned_budget=ibudget
        )
        testdb.add(t10)
        t11 = Transaction(
            date=(ppdate + timedelta(days=2)),
            budget_amounts={e2budget: Decimal('2.00')},
            description='prev trans 1',
            account=acct
        )
        testdb.add(t11)
        t12 = Transaction(
            date=(ppdate + timedelta(days=3)),
            budget_amounts={e1budget: Decimal('600.00')},
            budgeted_amount=Decimal('500.00'),
            description='prev trans 2',
            account=acct,
            planned_budget=e1budget
        )
        testdb.add(t12)
        testdb.flush()
        testdb.commit()

    def test_4_confirm_sums(self, testdb):
        periods = self.pay_periods(testdb)
        assert periods[0].overall_sums == {
            'allocated': Decimal('750.0'),
            'spent': Decimal('850.0'),
            'income': Decimal('1000.0'),
            'remaining': Decimal('150.0')
        }
        assert periods[1].overall_sums == {
            'allocated': Decimal('2350.0'),
            'spent': Decimal('2450.0'),
            'income': Decimal('1400.0'),
            'remaining': Decimal('-1050.0')
        }
        assert periods[2].overall_sums == {
            'allocated': Decimal('1288.0'),
            'spent': Decimal('1388.0'),
            'income': Decimal('1400.0'),
            'remaining': Decimal('12.0')
        }
        assert periods[3].overall_sums == {
            'allocated': Decimal('502.0'),
            'spent': Decimal('602.0'),
            'income': Decimal('1400.0'),
            'remaining': Decimal('798.0')
        }

    def test_5_pay_periods_table(self, base_url, selenium, testdb):
        periods = self.pay_periods(testdb)
        self.get(selenium, base_url + '/')
        table = selenium.find_element(By.ID, 'pay-period-table')
        texts = self.tbody2textlist(table)
        elems = self.tbody2elemlist(table)
        expected = [
            [
                periods[1].start_date.strftime('%Y-%m-%d') + ' (current)',
                '$2,350.00',
                '$2,450.00',
                '-$1,050.00'
            ],
            [
                periods[2].start_date.strftime('%Y-%m-%d'),
                '$1,288.00',
                '$1,388.00',
                '$12.00'
            ],
            [
                periods[3].start_date.strftime('%Y-%m-%d'),
                '$502.00',
                '$602.00',
                '$798.00'
            ]
        ]
        for i in range(4, 10):
            expected.append([
                periods[i].start_date.strftime('%Y-%m-%d'),
                '$500.00',
                '$0.00',
                '$500.00'
            ])
        assert texts == expected
        # test links
        links = [x[0].get_attribute('innerHTML') for x in elems]
        expected = []
        for idx, period in enumerate(periods):
            if idx == 0:
                continue
            dstr = period.start_date.strftime('%Y-%m-%d')
            s = '<a href="/payperiod/%s">%s</a>' % (dstr, dstr)
            if idx == 1:
                s += ' <em>(current)</em>'
            expected.append(s)
        assert links == expected
        # test red text for negative dollar amounts
        assert elems[0][3].get_attribute('innerHTML') == '<span ' \
            'class="text-danger">-$1,050.00</span>'
        # test highlighted row for current period
        tbody = table.find_element(By.TAG_NAME, 'tbody')
        trs = tbody.find_elements(By.TAG_NAME, 'tr')
        assert trs[0].get_attribute('class') == 'info'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestAccountTransfer(AcceptanceHelper):

    def test_01_verify_db(self, testdb):
        max_t = max([
            t.id for t in testdb.query(Transaction).all()
        ])
        assert max_t == 4
        accts = {
            t.id: t for t in testdb.query(Account).all()
        }
        assert accts[1].acct_type == AcctType.Bank
        assert accts[1].balance.ledger == Decimal('12789.01')
        assert accts[1].unreconciled_sum == Decimal('0.0')
        assert accts[2].acct_type == AcctType.Bank
        assert accts[2].balance.ledger == Decimal('100.23')
        assert accts[2].unreconciled_sum == Decimal('-333.33')
        assert accts[3].acct_type == AcctType.Credit
        assert accts[4].acct_type == AcctType.Credit
        assert accts[5].acct_type == AcctType.Investment
        assert accts[5].balance.ledger == Decimal('10362.91')

    def test_02_transfer_modal(self, base_url, selenium):
        # Fill in the form
        self.get(selenium, base_url + '/')
        # check the table content on the page
        btable = selenium.find_element(By.ID, 'table-accounts-bank')
        btexts = self.tbody2textlist(btable)
        assert btexts == [
            [
                'BankOne',
                '$12,789.01 (14 hours ago)',
                '$0.00',
                '$12,789.01'
            ],
            [
                'BankTwoStale',
                '$100.23 (a month ago)',
                '-$333.33',
                '$433.56'
            ]
        ]
        itable = selenium.find_element(By.ID, 'table-accounts-investment')
        itexts = self.tbody2textlist(itable)
        assert itexts == [
            [
                'InvestmentOne',
                '$10,362.91 (13 days ago)'
            ]
        ]
        # open the modal to do a transfer
        link = selenium.find_element(By.ID, 'btn_acct_txfr_bank')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Account Transfer'
        assert body.find_element(By.ID,
                                 'acct_txfr_frm_date').get_attribute('value') == dtnow(
            ).strftime('%Y-%m-%d')
        amt = body.find_element(By.ID, 'acct_txfr_frm_amount')
        amt.clear()
        amt.send_keys('123.45')
        budget_sel = Select(
            body.find_element(By.ID, 'acct_txfr_frm_budget')
        )
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['7', 'Income (i)']
        ]
        assert budget_sel.first_selected_option.get_attribute(
            'value') == 'None'
        budget_sel.select_by_value('2')
        from_acct_sel = Select(
            body.find_element(By.ID, 'acct_txfr_frm_from_account')
        )
        opts = []
        for o in from_acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert from_acct_sel.first_selected_option.get_attribute(
            'value'
        ) == 'None'
        from_acct_sel.select_by_value('1')
        to_acct_sel = Select(
            body.find_element(By.ID, 'acct_txfr_frm_to_account')
        )
        opts = []
        for o in to_acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert to_acct_sel.first_selected_option.get_attribute(
            'value'
        ) == 'None'
        to_acct_sel.select_by_value('2')
        notes = selenium.find_element(By.ID, 'acct_txfr_frm_notes')
        notes.clear()
        notes.send_keys('Account Transfer Notes')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transactions 5 and 6' \
                                 ' in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        self.wait_for_id(selenium, 'table-accounts-bank')
        # ensure that the page content updated after refreshing
        btable = selenium.find_element(By.ID, 'table-accounts-bank')
        btexts = self.tbody2textlist(btable)
        assert btexts == [
            [
                'BankOne',
                '$12,789.01 (14 hours ago)',
                '$123.45',
                '$12,665.56'
            ],
            [
                'BankTwoStale',
                '$100.23 (a month ago)',
                '-$456.78',
                '$557.01'
            ]
        ]
        itable = selenium.find_element(By.ID, 'table-accounts-investment')
        itexts = self.tbody2textlist(itable)
        assert itexts == [
            [
                'InvestmentOne',
                '$10,362.91 (13 days ago)'
            ]
        ]

    def test_03_verify_db(self, testdb):
        max_t = max([
            t.id for t in testdb.query(Transaction).all()
        ])
        assert max_t == 6
        desc = 'Account Transfer - 123.45 from BankOne (1) to BankTwoStale (2)'
        t2 = testdb.query(Transaction).get(5)
        assert t2.date == dtnow().date()
        assert t2.actual_amount == Decimal('123.45')
        assert t2.budgeted_amount == Decimal('123.45')
        assert t2.description == desc
        assert t2.notes == 'Account Transfer Notes'
        assert t2.account_id == 1
        assert t2.scheduled_trans_id is None
        assert len(t2.budget_transactions) == 1
        assert t2.budget_transactions[0].budget_id == 2
        assert t2.budget_transactions[0].amount == Decimal('123.45')
        t1 = testdb.query(Transaction).get(6)
        assert t1.date == dtnow().date()
        assert t1.actual_amount == Decimal('-123.45')
        assert t1.budgeted_amount == Decimal('-123.45')
        assert t1.description == desc
        assert t1.notes == 'Account Transfer Notes'
        assert t1.account_id == 2
        assert t1.scheduled_trans_id is None
        assert len(t1.budget_transactions) == 1
        assert t1.budget_transactions[0].budget_id == 2
        assert t1.budget_transactions[0].amount == Decimal('-123.45')
        acct1 = testdb.query(Account).get(1)
        assert acct1.balance.ledger == Decimal('12789.01')
        assert acct1.unreconciled_sum == Decimal('123.45')
        acct2 = testdb.query(Account).get(2)
        assert acct2.balance.ledger == Decimal('100.23')
        assert acct2.unreconciled_sum == Decimal('-456.78')



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from flask.views import MethodView
from flask import render_template, jsonify, request
from datatables import DataTable
from sqlalchemy import or_
import pickle
from base64 import b64decode

from biweeklybudget.flaskapp.app import app
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.models.account import Account
from biweeklybudget.db import db_session
from biweeklybudget.flaskapp.views.searchableajaxview import SearchableAjaxView
from biweeklybudget.ofxapi.local import OfxApiLocal
from biweeklybudget.ofxapi.exceptions import DuplicateFileException

logger = logging.getLogger(__name__)


class OfxView(MethodView):
    """
    Render the GET /ofx view using the ``ofx.html`` template.
    """

    def get(self):
        accts = {a.name: a.id for a in db_session.query(Account).all()}
        return render_template(
            'ofx.html',
            accts=accts
        )


class OfxTransView(MethodView):
    """
    Render the GET /ofx/<int:acct_id>/<str:fitid> view
    using the ``ofx.html`` template.
    """

    def get(self, acct_id, fitid):
        accts = {a.name: a.id for a in db_session.query(Account).all()}
        return render_template(
            'ofx.html',
            accts=accts,
            acct_id=acct_id,
            fitid=fitid
        )


class OfxTransAjax(MethodView):
    """
    Handle GET /ajax/ofx/<int:acct_id>/<str:fitid> endpoint.
    """

    def get(self, acct_id, fitid):
        txn = db_session.query(OFXTransaction).get((acct_id, fitid))
        stmt = txn.statement.as_dict
        aname = db_session.query(Account).get(acct_id).name
        res = {
            'acct_name': aname,
            'acct_id': txn.account_id,
            'txn': txn.as_dict,
            'stmt': stmt,
            'account_amount': txn.account_amount
        }
        res['txn']['account_name'] = aname
        res['txn']['account_amount'] = txn.account_amount
        return jsonify(res)


class OfxAccounts(MethodView):
    """
    Handle GET /api/ofx/accounts endpoint.

    This returns the JSON-ified return value from
    :py:meth:`~.OfxApiLocal.get_accounts` and will usually be called from
    :py:meth:`~.OfxApiRemote.get_accounts`.
    """

    def get(self):
        api = OfxApiLocal(db_session)
        return jsonify(api.get_accounts())


class OfxStatementPost(MethodView):
    """
    Handle POST /api/ofx/statement endpoint.

    This is a ReST API bridge between
    :py:meth:`~.OfxApiRemote.update_statement_ofx` on the client side and
    :py:meth:`~.OfxApiLocal.update_statement_ofx` on the server side.
    """

    def post(self):
        """
        Handle POST to /api/ofx/statement (from
        :py:meth:`~.OfxApiRemote.update_statement_ofx`) to upload a new OFX
        Statement (via :py:meth:`~.OfxApiLocal.update_statement_ofx`).

        The POSTed JSON should have the following keys:

        - ``acct_id`` (int) the Account ID the Statement is for
        - ``mtime`` (str) base64-encoded, pickled representation of the file
          modification time of the OFX file
        - ``filename`` (str) the file name of the OFX file
        - ``ofx`` (str) base64-encoded, pickled representation of the
          ``ofxparse.ofxparse.Ofx`` instance representing the Statement

        Returns a JSON object with the following fields:

        - ``success`` (bool) whether the operation was successful
        - ``message`` (str) message describing success or error message

        For successful operations, the JSON object will contain the following
        additional fields:

        - ``count_new`` (int) count of new transactions added
        - ``count_updated`` (int) count of transactions updated
        - ``statement_id`` (int) ID of the newly-added statement

        HTTP Status Codes:

        - 201 - Statement successfully added
        - 500 - DuplicateFileException
        - 400 - Any other error/exception
        """
        data = request.get_json()
        if sorted(data.keys()) != ['acct_id', 'filename', 'mtime', 'ofx']:
            logger.error('POST contained invalid or missing keys: %s',
                         data.keys())
            resp = jsonify({
                'success': False,
                'message': 'Invalid or missing JSON keys.'
            })
            resp.status_code = 400
            return resp
        try:
            ofx = pickle.loads(b64decode(data['ofx']))
            mtime = pickle.loads(b64decode(data['mtime']))
        except Exception:
            logger.error(
                'Error unpickling or base64-decoding OFX Statement post: %s',
                data['ofx'], exc_info=True
            )
            resp = jsonify({
                'success': False,
                'message': 'Unable to unpickle or b64decode "ofx" field.'
            })
            resp.status_code = 400
            return resp
        api = OfxApiLocal(db_session)
        try:
            stmt_id, count_new, count_upd = api.update_statement_ofx(
                data['acct_id'], ofx, mtime=mtime, filename=data['filename']
            )
        except DuplicateFileException as ex:
            resp = jsonify({
                'success': False,
                'message': 'File %s is a duplicate of stmt %d for account '
                           '%d' % (
                               ex.filename, ex.stmt_id, ex.acct_id
                           ),
                'account_id': ex.acct_id,
                'filename': ex.filename,
                'statement_id': ex.stmt_id
            })
            resp.status_code = 500
            return resp
        except Exception as ex:
            resp = jsonify({
                'success': False,
                'message': 'Exception: %s' % str(ex)
            })
            resp.status_code = 400
            return resp
        resp = jsonify({
            'success': True,
            'message': 'Successfully inserted Statement %d with %d new and %d '
                       'updated Transactions' % (
                           stmt_id, count_new, count_upd
                       ),
            'count_new': count_new,
            'count_updated': count_upd,
            'statement_id': stmt_id
        })
        resp.status_code = 201
        return resp


class OfxAjax(SearchableAjaxView):
    """
    Handle GET /ajax/ofx endpoint.
    """

    def _filterhack(self, qs, s, args):
        """
        DataTables 1.10.12 has built-in support for filtering based on a value
        in a specific column; when this is done, the filter value is set in
        ``columns[N][search][value]`` where N is the column number. However,
        the python datatables package used here only supports the global
        ``search[value]`` input, not the per-column one.

        However, the DataTable search is implemented by passing a callable
        to ``table.searchable()`` which takes two arguments, the current Query
        that's being built, and the user's ``search[value]`` input; this must
        then return a Query object with the search applied.

        In python datatables 0.4.9, this code path is triggered on
        ``if callable(self.search_func) and search.get("value", None):``

        As such, we can "trick" the table to use per-column searching (currently
        only if global searching is not being used) by examining the per-column
        search values in the request, and setting the search function to one
        (this method) that uses those values instead of the global
        ``search[value]``.

        :param qs: Query currently being built
        :type qs: ``sqlalchemy.orm.query.Query``
        :param s: user search value
        :type s: str
        :param args: args
        :type args: dict
        :return: Query with searching applied
        :rtype: ``sqlalchemy.orm.query.Query``
        """
        # Ok, build our filter...
        acct_filter = args['columns'][2]['search']['value']
        if acct_filter != '' and acct_filter != 'None':
            qs = qs.filter(OFXTransaction.account_id == acct_filter)
        # search
        if s != '' and s != 'FILTERHACK':
            if len(s) < 3:
                return qs
            s = '%' + s + '%'
            qs = qs.filter(or_(
                OFXTransaction.name.like(s),
                OFXTransaction.memo.like(s),
                OFXTransaction.description.like(s),
                OFXTransaction.notes.like(s)
            ))
        return qs

    def get(self):
        """
        Render and return JSON response for GET /ajax/ofx
        """
        args = request.args.to_dict()
        args_dict = self._args_dict(args)
        if self._have_column_search(args_dict) and args['search[value]'] == '':
            args['search[value]'] = 'FILTERHACK'
        table = DataTable(
            args, OFXTransaction, db_session.query(OFXTransaction), [
                (
                    'date',
                    'date_posted',
                    lambda i: i.date_posted.strftime('%Y-%m-%d')
                ),
                (
                    'amount',
                    lambda a: float(a.amount)
                ),
                (
                    'account',
                    'account.name',
                    lambda i: "{} ({})".format(i.name, i.id)
                ),
                ('type', 'trans_type'),
                'name',
                'memo',
                'description',
                'fitid',
                ('last_stmt', 'statement.id'),
                (
                    'last_stmt_date',
                    'statement.as_of',
                    lambda i: i.as_of.strftime('%Y-%m-%d')
                ),
                (
                    'reconcile_id',
                    'reconcile',
                    lambda i: None if i.reconcile is None else i.reconcile.id
                )
            ]
        )
        table.add_data(acct_id=lambda o: o.account_id)
        if args['search[value]'] != '':
            table.searchable(lambda qs, s: self._filterhack(qs, s, args_dict))
        return jsonify(table.json())


app.add_url_rule('/ofx', view_func=OfxView.as_view('ofx_view'))
app.add_url_rule('/ajax/ofx', view_func=OfxAjax.as_view('ofx_ajax'))
app.add_url_rule(
    '/ajax/ofx/<int:acct_id>/<fitid>',
    view_func=OfxTransAjax.as_view('ofx_trans_ajax')
)
app.add_url_rule(
    '/api/ofx/accounts',
    view_func=OfxAccounts.as_view('ofx_api_accounts')
)
app.add_url_rule(
    '/api/ofx/statement',
    view_func=OfxStatementPost.as_view('ofx_api_statement')
)
app.add_url_rule(
    '/ofx/<int:acct_id>/<fitid>',
    view_func=OfxTransView.as_view('ofx_trans')
)



"""Account remove re field defaults

Revision ID: 073142f641b3
Revises: 08b6358a04bf
Create Date: 2018-03-08 09:25:17.673039

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import mysql
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy import Column, Integer, String

Session = sessionmaker()

Base = declarative_base()

# revision identifiers, used by Alembic.
revision = '073142f641b3'
down_revision = '08b6358a04bf'
branch_labels = None
depends_on = None


class Account(Base):

    __tablename__ = 'accounts'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: regex for matching transactions as interest charges
    re_interest_charge = Column(String(254))

    #: regex for matching transactions as interest paid
    re_interest_paid = Column(String(254))

    #: regex for matching transactions as payments
    re_payment = Column(String(254))

    #: regex for matching transactions as late fees
    re_fee = Column(String(254))


def upgrade():
    bind = op.get_bind()
    session = Session(bind=bind)
    for acct in session.query(Account).all():
        acct.re_interest_charge = None
        acct.re_interest_paid = None
        acct.re_payment = None
        acct.re_fee = None
    session.commit()
    op.add_column(
        'accounts',
        sa.Column('re_late_fee', sa.String(length=254), nullable=True)
    )
    op.add_column(
        'accounts',
        sa.Column('re_other_fee', sa.String(length=254), nullable=True)
    )
    op.drop_column('accounts', 're_fee')


def downgrade():
    op.add_column(
        'accounts',
        sa.Column('re_fee', mysql.VARCHAR(length=254), nullable=True)
    )
    op.drop_column('accounts', 're_other_fee')
    op.drop_column('accounts', 're_late_fee')
    bind = op.get_bind()
    session = Session(bind=bind)
    for acct in session.query(Account).all():
        acct.re_interest_charge = '^(interest charge|purchase finance charge)'
        acct.re_interest_paid = '^interest paid'
        acct.re_payment = '^(online payment|' \
                          'internet payment|online pymt|payment)'
        acct.re_fee = '^(late fee|past due fee)'
    session.commit()



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from datetime import date
from decimal import Decimal

from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.interest import (
    InterestHelper, CCStatement, _BillingPeriod, AdbCompoundedDaily,
    MinPaymentAmEx, MinPaymentDiscover
)
from biweeklybudget.models import Account, OFXTransaction, OFXStatement
from biweeklybudget.utils import dtnow


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestInterestHelper(AcceptanceHelper):

    def test_00_setup_interest_charge_in_db(self, testdb):
        acct = testdb.query(Account).get(4)
        stmt = testdb.query(OFXStatement).get(7)
        txn = OFXTransaction(
            account=acct,
            statement=stmt,
            fitid='%s-MANUAL-CCPAYOFF' % dtnow().strftime('%Y%m%d%H%M%S'),
            trans_type='debit',
            date_posted=stmt.as_of,
            amount=Decimal('-46.9061'),
            name='Interest Charged - MANUALLY ENTERED',
            is_interest_charge=True
        )
        testdb.add(txn)
        testdb.commit()

    def test_01_init_accounts(self, testdb):
        cls = InterestHelper(testdb)
        assert cls._sess == testdb
        res = cls.accounts
        assert sorted(res.keys()) == [3, 4]
        assert isinstance(res[3], Account)
        assert res[3].id == 3
        assert isinstance(res[4], Account)
        assert res[4].id == 4

    def test_02_init_statements(self, testdb):
        cls = InterestHelper(testdb)
        res = cls._statements
        assert sorted(res.keys()) == [3, 4]
        s3 = res[3]
        assert isinstance(s3, CCStatement)
        assert isinstance(s3._billing_period, _BillingPeriod)
        assert s3._billing_period._end_date == date(2017, 7, 31)
        assert s3._billing_period._start_date == date(2017, 7, 1)
        assert isinstance(s3._interest_cls, AdbCompoundedDaily)
        assert s3._interest_cls.apr == Decimal('0.0100')
        assert isinstance(s3._min_pay_cls, MinPaymentAmEx)
        assert s3._orig_principal == Decimal('952.06')
        assert s3._min_pay is None
        assert s3._transactions == {}
        assert s3._principal == Decimal('952.06')
        s4 = res[4]
        assert isinstance(s4, CCStatement)
        assert isinstance(s4._billing_period, _BillingPeriod)
        assert s4._billing_period._end_date == date(2017, 7, 31)
        assert s4._billing_period._start_date == date(2017, 7, 1)
        assert isinstance(s4._interest_cls, AdbCompoundedDaily)
        assert s4._interest_cls.apr == Decimal('0.1000')
        assert isinstance(s4._min_pay_cls, MinPaymentDiscover)
        assert s4._orig_principal == Decimal('5498.65')
        assert s4._min_pay is None
        assert s4._transactions == {}
        assert s4._principal == Decimal('5498.65')

    def test_03_min_payments(self, testdb):
        cls = InterestHelper(testdb)
        assert cls.min_payments == {
            3: Decimal('35'),
            4: Decimal('109.9730')
        }

    def test_04_payoffs_simple(self, testdb):
        cls = InterestHelper(testdb)
        assert cls.calculate_payoffs() == {
            'HighestBalanceFirstMethod': {
                'description': 'Highest to Lowest Balance',
                'doc': 'Pay statements off from highest to lowest balance.',
                'results': {
                    3: {
                        'payoff_months': 28,
                        'total_interest': Decimal(
                            '10.9388625702411101133192793'
                        ),
                        'total_payments': Decimal(
                            '962.9988625702411101133192793'
                        ),
                        'next_payment': Decimal('35')
                    },
                    4: {
                        'payoff_months': 55,
                        'total_interest': Decimal(
                            '1457.695228060182432444990377'
                        ),
                        'total_payments': Decimal(
                            '6956.345228060182432444990377'
                        ),
                        'next_payment': Decimal('109.9730')
                    }
                }
            },
            'HighestInterestRateFirstMethod': {
                'description': 'Highest to Lowest Interest Rate',
                'doc': 'Pay statements off from highest to lowest interest '
                       'rate.',
                'results': {
                    3: {
                        'payoff_months': 28,
                        'total_interest': Decimal(
                            '10.9388625702411101133192793'
                        ),
                        'total_payments': Decimal(
                            '962.9988625702411101133192793'
                        ),
                        'next_payment': Decimal('35')
                    },
                    4: {
                        'payoff_months': 55,
                        'total_interest': Decimal(
                            '1457.695228060182432444990377'
                        ),
                        'total_payments': Decimal(
                            '6956.345228060182432444990377'
                        ),
                        'next_payment': Decimal('109.9730')
                    }
                }
            },
            'LowestBalanceFirstMethod': {
                'description': 'Lowest to Highest Balance (a.k.a. '
                               'Snowball Method)',
                'doc': 'Pay statements off from lowest to highest balance, '
                       'a.k.a. the "snowball"\nmethod.',
                'results': {
                    3: {
                        'payoff_months': 21,
                        'total_interest': Decimal(
                            '8.8578327498502165965138131'
                        ),
                        'total_payments': Decimal(
                            '960.9178327498502165965138131'
                        ),
                        'next_payment': Decimal('35')
                    },
                    4: {
                        'payoff_months': 56,
                        'total_interest': Decimal(
                            '1489.587124948955044765363412'
                        ),
                        'total_payments': Decimal(
                            '6988.237124948955044765363412'
                        ),
                        'next_payment': Decimal('109.9730')
                    }
                }
            },
            'LowestInterestRateFirstMethod': {
                'description': 'Lowest to Highest Interest Rate',
                'doc': 'Pay statements off from lowest to highest '
                       'interest rate.',
                'results': {
                    3: {
                        'payoff_months': 21,
                        'total_interest': Decimal(
                            '8.8578327498502165965138131'
                        ),
                        'total_payments': Decimal(
                            '960.9178327498502165965138131'
                        ),
                        'next_payment': Decimal('35')
                    },
                    4: {
                        'payoff_months': 56,
                        'total_interest': Decimal(
                            '1489.587124948955044765363412'
                        ),
                        'total_payments': Decimal(
                            '6988.237124948955044765363412'
                        ),
                        'next_payment': Decimal('109.9730')
                    }
                }
            },
            'MinPaymentMethod': {
                'description': 'Minimum Payment Only',
                'doc': 'Pay only the minimum on each statement.',
                'results': {
                    3: {
                        'payoff_months': 28,
                        'total_interest': Decimal(
                            '10.9388625702411101133192793'
                        ),
                        'total_payments': Decimal(
                            '962.9988625702411101133192793'
                        ),
                        'next_payment': Decimal('35')
                    },
                    4: {
                        'payoff_months': 162,
                        'total_interest': Decimal(
                            '3166.211877369277471400473622'
                        ),
                        'total_payments': Decimal(
                            '8664.861877369277471400473622'
                        ),
                        'next_payment': Decimal('109.9730')
                    }
                }
            }
        }



from __future__ import absolute_import
from __future__ import unicode_literals
import datetime
import hashlib
try:
    # python 3
    from io import StringIO, BytesIO
    IS_PYTHON_2 = False
except ImportError:
    # python 2
    from StringIO import StringIO
    IS_PYTHON_2 = True
import time

from ofxparse import OfxParser, AccountType


class Account(object):
    """Base class for accounts at an institution

    :param number: The account number
    :type number: string
    :param institution: The bank this belongs to
    :type institution: :py:class:`ofxclient.Institution` object
    :param description: optional account description
    :type description: string or None

    This class is almost never never instantiated on it's own. Instead,
    sub-classes are instantiated.

    In most cases these subclasses are either being deserialized from a
    config file entry, a serialization hash, or returned by the
    :py:meth:`ofxclient.Institution.accounts` method.

    Example from a saved config entry::

      from biweeklybudget.vendored.ofxclient.config import OfxConfig
      account = OfxConfig().account('local_id() string')

    Example of deserialization::

      from biweeklybudget.vendored.ofxclient import BankAccount
      # assume 'inst' is an Institution()
      a1    = BankAccount(number='asdf',institution=inst)
      data1 = a1.serialize()
      a2    = Account.deserialize(data1)

    Example by querying the bank directly::

      from biweeklybudget.vendored.ofxclient import Institution
      # assume an Institution() is configured with
      # a username/password etc
      accounts = institution.accounts()

    .. seealso::

       :py:class:`ofxclient.BankAccount`
       :py:class:`ofxclient.BrokerageAccount`
       :py:class:`ofxclient.CreditCardAccount`

    """
    def __init__(self, number, institution, description=None):
        self.institution = institution
        self.number = number
        self.description = description or self._default_description()

    def local_id(self):
        """Locally generated unique account identifier.

        :rtype: string
        """
        return hashlib.sha256(("%s%s" % (
            self.institution.local_id(),
            self.number)).encode()).hexdigest()

    def number_masked(self):
        """Masked version of the account number for privacy.

        :rtype: string
        """
        return "***%s" % self.number[-4:]

    def long_description(self):
        """Long description of the account (includes institution description).

        :rtype: string
        """
        return "%s: %s" % (self.institution.description, self.description)

    def _default_description(self):
        return self.number_masked()

    def download(self, days=60):
        """Downloaded OFX response for the given time range

        :param days: Number of days to look back at
        :type days: integer
        :rtype: :py:class:`StringIO`

        """
        days_ago = datetime.datetime.now() - datetime.timedelta(days=days)
        as_of = time.strftime("%Y%m%d", days_ago.timetuple())
        query = self._download_query(as_of=as_of)
        response = self.institution.client().post(query)
        return StringIO(response)

    def download_parsed(self, days=60):
        """Downloaded OFX response parsed by :py:meth:`OfxParser.parse`

        :param days: Number of days to look back at
        :type days: integer
        :rtype: :py:class:`ofxparser.Ofx`
        """
        if IS_PYTHON_2:
            return OfxParser.parse(
                self.download(days=days)
            )
        else:
            return OfxParser.parse(
                BytesIO(self.download(days=days).read().encode())
            )

    def statement(self, days=60):
        """Download the :py:class:`ofxparse.Statement` given the time range

        :param days: Number of days to look back at
        :type days: integer
        :rtype: :py:class:`ofxparser.Statement`
        """
        parsed = self.download_parsed(days=days)
        return parsed.account.statement

    def transactions(self, days=60):
        """Download a a list of :py:class:`ofxparse.Transaction` objects

        :param days: Number of days to look back at
        :type days: integer
        :rtype: list of :py:class:`ofxparser.Transaction` objects
        """
        return self.statement(days=days).transactions

    def serialize(self):
        """Serialize predictably for use in configuration storage.

        Output look like this::

          {
            'local_id':       'string',
            'number':         'account num',
            'description':    'descr',
            'broker_id':      'may be missing - type dependent',
            'routing_number': 'may be missing - type dependent,
            'account_type':   'may be missing - type dependent,
            'institution': {
                # ... see :py:meth:`ofxclient.Institution.serialize`
            }
          }

        :rtype: nested dictionary
        """
        data = {
            'local_id': self.local_id(),
            'institution': self.institution.serialize(),
            'number': self.number,
            'description': self.description
        }
        if hasattr(self, 'broker_id'):
            data['broker_id'] = self.broker_id
        elif hasattr(self, 'routing_number'):
            data['routing_number'] = self.routing_number
            data['account_type'] = self.account_type

        return data

    @staticmethod
    def deserialize(raw):
        """Instantiate :py:class:`ofxclient.Account` subclass from dictionary

        :param raw: serilized Account
        :param type: dict as  given by :py:meth:`~ofxclient.Account.serialize`
        :rtype: subclass of :py:class:`ofxclient.Account`
        """
        from biweeklybudget.vendored.ofxclient.institution import Institution
        institution = Institution.deserialize(raw['institution'])

        del raw['institution']
        del raw['local_id']

        if 'broker_id' in raw:
            a = BrokerageAccount(institution=institution, **raw)
        elif 'routing_number' in raw:
            a = BankAccount(institution=institution, **raw)
        else:
            a = CreditCardAccount(institution=institution, **raw)
        return a

    @staticmethod
    def from_ofxparse(data, institution):
        """Instantiate :py:class:`ofxclient.Account` subclass from ofxparse
        module

        :param data: an ofxparse account
        :type data: An :py:class:`ofxparse.Account` object
        :param institution: The parent institution of the account
        :type institution: :py:class:`ofxclient.Institution` object
        """

        description = data.desc if hasattr(data, 'desc') else None
        if data.type == AccountType.Bank:
            return BankAccount(
                institution=institution,
                number=data.account_id,
                routing_number=data.routing_number,
                account_type=data.account_type,
                description=description)
        elif data.type == AccountType.CreditCard:
            return CreditCardAccount(
                institution=institution,
                number=data.account_id,
                description=description)
        elif data.type == AccountType.Investment:
            return BrokerageAccount(
                institution=institution,
                number=data.account_id,
                broker_id=data.brokerid,
                description=description)
        raise ValueError("unknown account type: %s" % data.type)


class BrokerageAccount(Account):
    """:py:class:`ofxclient.Account` subclass for brokerage/investment accounts

    In addition to the parameters it's superclass requires, the following
    parameters are needed.

    :param broker_id: Broker ID of the account
    :type broker_id: string

    .. seealso::

       :py:class:`ofxclient.Account`
    """
    def __init__(self, broker_id, **kwargs):
        super(BrokerageAccount, self).__init__(**kwargs)
        self.broker_id = broker_id

    def _download_query(self, as_of):
        """Formulate the specific query needed for download

        Not intended to be called by developers directly.

        :param as_of: Date in 'YYYYMMDD' format
        :type as_of: string
        """
        c = self.institution.client()
        q = c.brokerage_account_query(
            number=self.number, date=as_of, broker_id=self.broker_id)
        return q


class BankAccount(Account):
    """:py:class:`ofxclient.Account` subclass for a checking/savings account

    In addition to the parameters it's superclass requires, the following
    parameters are needed.

    :param routing_number: Routing number or account number of the account
    :type routing_number: string
    :param account_type: Account type per OFX spec can be empty but not None
    :type account_type: string

    .. seealso::

       :py:class:`ofxclient.Account`
    """
    def __init__(self, routing_number, account_type, **kwargs):
        super(BankAccount, self).__init__(**kwargs)
        self.routing_number = routing_number
        self.account_type = account_type

    def _download_query(self, as_of):
        """Formulate the specific query needed for download

        Not intended to be called by developers directly.

        :param as_of: Date in 'YYYYMMDD' format
        :type as_of: string
        """
        c = self.institution.client()
        q = c.bank_account_query(
            number=self.number,
            date=as_of,
            account_type=self.account_type,
            bank_id=self.routing_number)
        return q


class CreditCardAccount(Account):
    """:py:class:`ofxclient.Account` subclass for a credit card account

    No additional parameters to the constructor are needed.

    .. seealso::

       :py:class:`ofxclient.Account`
    """
    def __init__(self, **kwargs):
        super(CreditCardAccount, self).__init__(**kwargs)

    def _download_query(self, as_of):
        """Formulate the specific query needed for download

        Not intended to be called by developers directly.

        :param as_of: Date in 'YYYYMMDD' format
        :type as_of: string
        """
        c = self.institution.client()
        q = c.credit_card_account_query(number=self.number, date=as_of)
        return q



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from biweeklybudget.utils import dtnow, plaid_client
from pytz import utc
from datetime import datetime
from freezegun import freeze_time

from unittest.mock import patch, call, Mock, DEFAULT

pbm: str = 'biweeklybudget.utils'


class TestPlaidClient:

    def test_happy_path(self):
        mock_client = Mock()
        mock_config = Mock()
        mock_api = Mock()

        class MockEnv:
            Production = 'prodEnv'
            Development = 'devEnv'
            Sandbox = 'sandEnv'

        with patch.multiple(
            pbm,
            Configuration=DEFAULT,
            ApiClient=DEFAULT,
            Environment=MockEnv,
            PlaidApi=DEFAULT
        ) as mocks:
            mocks['Configuration'].return_value = mock_config
            mocks['ApiClient'].return_value = mock_client
            mocks['PlaidApi'].return_value = mock_api
            res = plaid_client()
        assert mocks['Configuration'].mock_calls == [
            call(
                host='sandEnv',
                api_key={
                    'clientId': 'plaidCID',
                    'secret': 'plaidSecret'
                }
            )
        ]
        assert mocks['ApiClient'].mock_calls == [
            call(mock_config)
        ]
        assert mocks['PlaidApi'].mock_calls == [
            call(mock_client)
        ]
        assert res == mock_api


class TestDtNow(object):

    @freeze_time('2016-05-13 11:21:32', tz_offset=0)
    @patch('biweeklybudget.utils.settings.BIWEEKLYBUDGET_TEST_TIMESTAMP', None)
    def test_dtnow(self):
        assert dtnow() == datetime(
            2016, 5, 13, 11, 21, 32, tzinfo=utc
        )

    def test_dtnow_test(self):
        assert dtnow() == datetime(
            2017, 7, 28, 6, 24, 44, tzinfo=utc
        )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from datetime import timedelta, date
from selenium.webdriver.support.ui import Select
from selenium.webdriver.common.by import By
from decimal import Decimal
from time import sleep

from biweeklybudget.utils import dtnow
from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.models.scheduled_transaction import ScheduledTransaction


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestSchedTrans(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/scheduled')

    def test_heading(self, selenium):
        heading = selenium.find_element(By.CLASS_NAME, 'navbar-brand')
        assert heading.text == 'Scheduled Transactions - BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element(By.ID, 'side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element(By.ID, 'notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestSchedTransDefault(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.dt = dtnow()
        self.get(selenium, base_url + '/scheduled')

    def test_table(self, selenium):
        table = selenium.find_element(By.ID, 'table-scheduled-txn')
        texts = self.tbody2textlist(table)
        # Extract just the descriptions and types for verification
        descriptions = [t[4] for t in texts]
        # Verify all expected transactions are present
        expected_descriptions = ['ST1', 'ST2', 'ST3', 'ST4', 'ST5', 'ST6']
        for desc in expected_descriptions:
            assert desc in descriptions, f'{desc} not found in table'

    def test_filter_opts(self, selenium):
        self.get(selenium, self.baseurl + '/scheduled')
        acct_filter = Select(selenium.find_element(By.ID, 'type_filter'))
        # find the options
        opts = []
        for o in acct_filter.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['date', 'Date'],
            ['monthly', 'Monthly'],
            ['per period', 'Per Period'],
            ['weekly', 'Weekly'],
            ['annual', 'Annual']
        ]

    def test_filter(self, selenium):
        self.get(selenium, self.baseurl + '/scheduled')
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-scheduled-txn')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[4] for t in texts]
        # check all 6 transactions are present
        assert len(trans) == 6
        type_filter = Select(selenium.find_element(By.ID, 'type_filter'))
        # select Monthly
        type_filter.select_by_value('monthly')
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-scheduled-txn')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[4] for t in texts]
        assert set(trans) == {'ST2', 'ST5'}
        # select back to all
        type_filter.select_by_value('None')
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-scheduled-txn')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[4] for t in texts]
        assert len(trans) == 6

    def test_search(self, selenium):
        self.get(selenium, self.baseurl + '/scheduled')
        search = self.retry_stale(
            lambda: selenium.find_element(By.XPATH, '//input[@type="search"]')
        )
        search.send_keys('ST3')
        sleep(3)  # yeah, yeah, I know...
        table = self.retry_stale(
            lambda: selenium.find_element(By.ID, 'table-scheduled-txn')
        )
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [t[4] for t in texts]
        # check sanity
        assert trans == ['ST3']


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestSchedTransModalPerPeriod(AcceptanceHelper):

    def test_0_verify_db(self, testdb):
        t = testdb.query(ScheduledTransaction).get(3)
        assert t is not None
        assert t.description == 'ST3'
        assert t.num_per_period == 1
        assert t.date is None
        assert t.day_of_month is None
        assert t.amount == Decimal('-333.33')
        assert t.account_id == 2
        assert t.budget_id == 4
        assert t.notes == 'notesST3'
        assert t.is_active is True

    def test_1_modal_on_click(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/scheduled')
        link = selenium.find_element(By.XPATH, '//a[text()="ST3"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Scheduled Transaction 3'
        assert body.find_element(By.ID,
                                 'sched_frm_id').get_attribute('value') == '3'
        assert body.find_element(By.ID,
                                 'sched_frm_description').get_attribute('value') == 'ST3'
        assert body.find_element(By.ID,
                                 'sched_frm_type_monthly').is_selected() is False
        assert body.find_element(By.ID,
                                 'sched_frm_type_date').is_selected() is False
        assert body.find_element(By.ID,
                                 'sched_frm_type_per_period').is_selected()
        assert body.find_element(By.ID,
                                 'sched_frm_num_per_period').get_attribute('value') == '1'
        assert body.find_element(By.ID,
                                 'sched_frm_amount').get_attribute('value') == '-333.33'
        acct_sel = Select(body.find_element(By.ID, 'sched_frm_account'))
        opts = []
        for o in acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert acct_sel.first_selected_option.get_attribute('value') == '2'
        budget_sel = Select(body.find_element(By.ID, 'sched_frm_budget'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['7', 'Income (income)'],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['3', 'Periodic3 Inactive'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['6', 'Standing3 Inactive']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '4'
        assert selenium.find_element(By.ID, 'sched_frm_active').is_selected()


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestSchedTransMonthlyURL(AcceptanceHelper):

    def test_0_verify_db(self, testdb):
        t = testdb.query(ScheduledTransaction).get(2)
        assert t is not None
        assert t.description == 'ST2'
        assert t.num_per_period is None
        assert t.date is None
        assert t.day_of_month == 4
        assert t.amount == Decimal('222.22')
        assert t.account_id == 1
        assert t.budget_id == 2
        assert t.notes == 'notesST2'
        assert t.is_active is True

    def test_1_modal_from_url(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/scheduled/2')
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Scheduled Transaction 2'
        assert body.find_element(By.ID,
                                 'sched_frm_id').get_attribute('value') == '2'
        assert body.find_element(By.ID,
                                 'sched_frm_description').get_attribute('value') == 'ST2'
        assert body.find_element(By.ID,
                                 'sched_frm_type_monthly').is_selected()
        assert body.find_element(By.ID,
                                 'sched_frm_type_date').is_selected() is False
        assert body.find_element(By.ID,
                                 'sched_frm_type_per_period').is_selected() is False
        assert body.find_element(By.ID,
                                 'sched_frm_day_of_month').get_attribute('value') == '4'
        assert body.find_element(By.ID,
                                 'sched_frm_amount').get_attribute('value') == '222.22'
        acct_sel = Select(body.find_element(By.ID, 'sched_frm_account'))
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        budget_sel = Select(body.find_element(By.ID, 'sched_frm_budget'))
        assert budget_sel.first_selected_option.get_attribute('value') == '2'
        assert selenium.find_element(By.ID, 'sched_frm_active').is_selected()


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.incremental
class TestSchedTransModal(AcceptanceHelper):

    def test_00_edit_date_inactive_verify_db(self, testdb):
        t = testdb.query(ScheduledTransaction).get(4)
        assert t is not None
        assert t.description == 'ST4'
        assert t.num_per_period is None
        assert t.date == (
            dtnow() + timedelta(days=5)
        ).date()
        assert t.day_of_month is None
        assert t.amount == Decimal('444.44')
        assert t.account_id == 1
        assert t.budget_id == 1
        assert t.notes == 'notesST4'
        assert t.is_active is False

    def test_01_edit_date_inactive_modal_from_url(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/scheduled/4')
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Scheduled Transaction 4'
        assert body.find_element(By.ID,
                                 'sched_frm_id').get_attribute('value') == '4'
        assert body.find_element(By.ID,
                                 'sched_frm_description').get_attribute('value') == 'ST4'
        assert body.find_element(By.ID,
                                 'sched_frm_type_monthly').is_selected() is False
        assert body.find_element(By.ID,
                                 'sched_frm_type_date').is_selected()
        assert body.find_element(By.ID,
                                 'sched_frm_type_per_period').is_selected() is False
        assert body.find_element(By.ID,
                                 'sched_frm_date').get_attribute('value') == (
            dtnow() + timedelta(days=5)).strftime('%Y-%m-%d')
        assert body.find_element(By.ID,
                                 'sched_frm_amount').get_attribute('value') == '444.44'
        acct_sel = Select(body.find_element(By.ID, 'sched_frm_account'))
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        budget_sel = Select(body.find_element(By.ID, 'sched_frm_budget'))
        assert budget_sel.first_selected_option.get_attribute('value') == '1'
        assert selenium.find_element(By.ID,
                                     'sched_frm_active').is_selected() is False

    def test_02_edit_date_inactive_modal_edit(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/scheduled/4')
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        desc = body.find_element(By.ID, 'sched_frm_description')
        desc.send_keys('edited')
        _type = body.find_element(By.ID, 'sched_frm_type_date')
        _type.click()
        date_input = body.find_element(By.ID, 'sched_frm_date')
        date_input.clear()
        date_input.send_keys((dtnow() + timedelta(days=1)).strftime('%Y-%m-%d'))
        amt = body.find_element(By.ID, 'sched_frm_amount')
        amt.clear()
        amt.send_keys('123.45')
        acct_sel = Select(body.find_element(By.ID, 'sched_frm_account'))
        acct_sel.select_by_value('2')
        budget_sel = Select(body.find_element(By.ID, 'sched_frm_budget'))
        budget_sel.select_by_value('4')
        is_active = selenium.find_element(By.ID, 'sched_frm_active')
        is_active.click()
        assert is_active.is_selected()
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved ScheduledTransaction 4 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that updated budget was removed from the page
        table = selenium.find_element(By.ID, 'table-scheduled-txn')
        texts = self.tbody2textlist(table)
        # sort order changes when we make this active
        assert texts[2][0] == 'yes'
        assert texts[2][4] == 'ST4edited'

    def test_03_edit_date_inactive_verify_db(self, testdb):
        t = testdb.query(ScheduledTransaction).get(4)
        assert t is not None
        assert t.description == 'ST4edited'
        assert t.num_per_period is None
        assert t.date == (
            dtnow() + timedelta(days=1)
        ).date()
        assert t.day_of_month is None
        assert t.amount == Decimal('123.45')
        assert t.account_id == 2
        assert t.budget_id == 4
        assert t.notes == 'notesST4'
        assert t.is_active is True

    def test_10_add_date_modal_on_click(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/scheduled')
        link = selenium.find_element(By.ID, 'btn_add_sched')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Scheduled Transaction'
        desc = body.find_element(By.ID, 'sched_frm_description')
        desc.send_keys('NewST7')
        _type = body.find_element(By.ID, 'sched_frm_type_date')
        _type.click()
        date_input = body.find_element(By.ID, 'sched_frm_date')
        assert date_input.is_displayed()
        # BEGIN select the 15th of this month from the popup
        dnow = dtnow()
        expected_date = date(year=dnow.year, month=dnow.month, day=15)
        date_input.click()
        date_number = body.find_element(By.XPATH,
                                        '//td[@class="day" and text()="15"]'
                                        )
        date_number.click()
        # END date chooser popup
        assert date_input.get_attribute(
            'value') == expected_date.strftime('%Y-%m-%d')
        amt = body.find_element(By.ID, 'sched_frm_amount')
        amt.send_keys('123.45')
        acct_sel = Select(body.find_element(By.ID, 'sched_frm_account'))
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        acct_sel.select_by_value('1')
        budget_sel = Select(body.find_element(By.ID, 'sched_frm_budget'))
        budget_sel.select_by_value('1')
        is_active = selenium.find_element(By.ID, 'sched_frm_active')
        assert is_active.is_selected()
        notes = body.find_element(By.ID, 'sched_frm_notes')
        notes.send_keys('foo bar baz')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved ScheduledTransaction 7 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that updated budget was removed from the page
        table = selenium.find_element(By.ID, 'table-scheduled-txn')
        texts = [y[4] for y in self.tbody2textlist(table)]
        assert 'NewST7' in texts

    def test_13_add_date_verify_db(self, testdb):
        t = testdb.query(ScheduledTransaction).get(7)
        assert t is not None
        assert t.description == 'NewST7'
        assert t.num_per_period is None
        dnow = dtnow()
        assert t.date == date(year=dnow.year, month=dnow.month, day=15)
        assert t.day_of_month is None
        assert t.amount == Decimal('123.45')
        assert t.account_id == 1
        assert t.budget_id == 1
        assert t.notes == 'foo bar baz'
        assert t.is_active is True

    def test_21_add_monthly_modal_on_click(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/scheduled')
        link = selenium.find_element(By.ID, 'btn_add_sched')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Scheduled Transaction'
        desc = body.find_element(By.ID, 'sched_frm_description')
        desc.send_keys('NewST8Monthly')
        _type = body.find_element(By.ID, 'sched_frm_type_monthly')
        _type.click()
        day_input = body.find_element(By.ID, 'sched_frm_day_of_month')
        assert day_input.is_displayed()
        day_input.send_keys('4')
        amt = body.find_element(By.ID, 'sched_frm_amount')
        amt.send_keys('123.45')
        acct_sel = Select(body.find_element(By.ID, 'sched_frm_account'))
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        acct_sel.select_by_value('2')
        budget_sel = Select(body.find_element(By.ID, 'sched_frm_budget'))
        budget_sel.select_by_value('2')
        is_active = selenium.find_element(By.ID, 'sched_frm_active')
        assert is_active.is_selected()
        notes = body.find_element(By.ID, 'sched_frm_notes')
        notes.send_keys('foo bar baz')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved ScheduledTransaction 8 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that updated budget was removed from the page
        table = selenium.find_element(By.ID, 'table-scheduled-txn')
        texts = [y[4] for y in self.tbody2textlist(table)]
        assert 'NewST8Monthly' in texts

    def test_23_add_monthly_verify_db(self, testdb):
        t = testdb.query(ScheduledTransaction).get(8)
        assert t is not None
        assert t.description == 'NewST8Monthly'
        assert t.num_per_period is None
        assert t.date is None
        assert t.day_of_month == 4
        assert t.amount == Decimal('123.45')
        assert t.account_id == 2
        assert t.budget_id == 2
        assert t.notes == 'foo bar baz'
        assert t.is_active is True

    def test_31_add_per_period_modal_on_click(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/scheduled')
        link = selenium.find_element(By.ID, 'btn_add_sched')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Scheduled Transaction'
        desc = body.find_element(By.ID, 'sched_frm_description')
        desc.send_keys('NewST9PerPeriod')
        _type = body.find_element(By.ID, 'sched_frm_type_per_period')
        _type.click()
        date_input = body.find_element(By.ID, 'sched_frm_num_per_period')
        assert date_input.is_displayed()
        date_input.send_keys('2')
        amt = body.find_element(By.ID, 'sched_frm_amount')
        amt.send_keys('123.45')
        acct_sel = Select(body.find_element(By.ID, 'sched_frm_account'))
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        acct_sel.select_by_value('1')
        budget_sel = Select(body.find_element(By.ID, 'sched_frm_budget'))
        budget_sel.select_by_value('1')
        is_active = selenium.find_element(By.ID, 'sched_frm_active')
        assert is_active.is_selected()
        notes = body.find_element(By.ID, 'sched_frm_notes')
        notes.send_keys('foo bar baz')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved ScheduledTransaction 9 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that updated budget was removed from the page
        table = selenium.find_element(By.ID, 'table-scheduled-txn')
        texts = [y[4] for y in self.tbody2textlist(table)]
        assert 'NewST9PerPeriod' in texts

    def test_33_add_per_period_verify_db(self, testdb):
        t = testdb.query(ScheduledTransaction).get(9)
        assert t is not None
        assert t.description == 'NewST9PerPeriod'
        assert t.num_per_period == 2
        assert t.date is None
        assert t.day_of_month is None
        assert t.amount == Decimal('123.45')
        assert t.account_id == 1
        assert t.budget_id == 1
        assert t.notes == 'foo bar baz'
        assert t.is_active is True

    def test_41_add_income_modal_on_click(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/scheduled')
        link = selenium.find_element(By.ID, 'btn_add_sched')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Scheduled Transaction'
        desc = body.find_element(By.ID, 'sched_frm_description')
        desc.send_keys('NewST10PerPeriod')
        _type = body.find_element(By.ID, 'sched_frm_type_per_period')
        _type.click()
        date_input = body.find_element(By.ID, 'sched_frm_num_per_period')
        assert date_input.is_displayed()
        date_input.send_keys('1')
        amt = body.find_element(By.ID, 'sched_frm_amount')
        amt.send_keys('123.45')
        acct_sel = Select(body.find_element(By.ID, 'sched_frm_account'))
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        acct_sel.select_by_value('1')
        budget_sel = Select(body.find_element(By.ID, 'sched_frm_budget'))
        budget_sel.select_by_value('7')
        is_active = selenium.find_element(By.ID, 'sched_frm_active')
        assert is_active.is_selected()
        notes = body.find_element(By.ID, 'sched_frm_notes')
        notes.send_keys('foo bar baz')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved ScheduledTransaction 10 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that updated budget was removed from the page
        table = selenium.find_element(By.ID, 'table-scheduled-txn')
        texts = [y[4] for y in self.tbody2textlist(table)]
        assert 'NewST10PerPeriod' in texts

    def test_43_add_income_verify_db(self, testdb):
        t = testdb.query(ScheduledTransaction).get(10)
        assert t is not None
        assert t.description == 'NewST10PerPeriod'
        assert t.num_per_period == 1
        assert t.date is None
        assert t.day_of_month is None
        assert t.amount == Decimal('123.45')
        assert t.account_id == 1
        assert t.budget_id == 7
        assert t.notes == 'foo bar baz'
        assert t.is_active is True

    def test_44_add_income_table(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/scheduled')
        table = selenium.find_element(By.ID, 'table-scheduled-txn')
        texts = self.tbody2textlist(table)
        elems = self.tbody2elemlist(table)
        assert texts[1] == [
            'yes',
            'per period',
            '1 per period',
            '$123.45',
            'NewST10PerPeriod',
            'BankOne (1)',
            'Income (income) (7)'
        ]
        assert elems[1][4].get_attribute('innerHTML') == '' \
            '<a href="javascript:schedModal(10, mytable)">NewST10PerPeriod</a>'
        assert elems[1][5].get_attribute('innerHTML') == '' \
            '<a href="/accounts/1">BankOne (1)</a>'
        assert elems[1][6].get_attribute('innerHTML') == '' \
            '<a href="/budgets/7">Income (income) (7)</a>'

    def test_50_sales_tax_verify_db(self, testdb):
        """Verify initial sales_tax values in database"""
        t = testdb.query(ScheduledTransaction).get(1)
        assert t is not None
        assert t.description == 'ST1'
        assert t.sales_tax == Decimal('1.23')

    def test_51_sales_tax_edit_modal(self, base_url, selenium):
        """Test editing sales_tax value in modal"""
        self.baseurl = base_url
        self.get(selenium, base_url + '/scheduled/1')
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Scheduled Transaction 1'
        # verify sales_tax field is present and populated
        sales_tax_elem = body.find_element(By.ID, 'sched_frm_sales_tax')
        assert sales_tax_elem.get_attribute('value') == '1.23'
        # change sales_tax value
        sales_tax_elem.clear()
        sales_tax_elem.send_keys('9.87')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved ScheduledTransaction 1 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)

    def test_52_sales_tax_verify_db_after_edit(self, testdb):
        """Verify sales_tax was updated in database"""
        t = testdb.query(ScheduledTransaction).get(1)
        assert t is not None
        assert t.description == 'ST1'
        assert t.sales_tax == Decimal('9.87')

    def test_53_sales_tax_add_with_sales_tax(self, base_url, selenium):
        """Test adding new scheduled transaction with sales_tax"""
        self.baseurl = base_url
        self.get(selenium, base_url + '/scheduled')
        link = selenium.find_element(By.ID, 'btn_add_sched')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Scheduled Transaction'
        desc = body.find_element(By.ID, 'sched_frm_description')
        desc.send_keys('NewSTWithSalesTax')
        _type = body.find_element(By.ID, 'sched_frm_type_monthly')
        _type.click()
        dom = body.find_element(By.ID, 'sched_frm_day_of_month')
        dom.send_keys('15')
        amt = body.find_element(By.ID, 'sched_frm_amount')
        amt.send_keys('250.00')
        sales_tax = body.find_element(By.ID, 'sched_frm_sales_tax')
        sales_tax.send_keys('12.50')
        acct_sel = Select(body.find_element(By.ID, 'sched_frm_account'))
        acct_sel.select_by_value('1')
        budget_sel = Select(body.find_element(By.ID, 'sched_frm_budget'))
        budget_sel.select_by_value('1')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved ScheduledTransaction 11 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)

    def test_54_sales_tax_verify_db_after_add(self, testdb):
        """Verify new scheduled transaction with sales_tax in database"""
        t = testdb.query(ScheduledTransaction).get(11)
        assert t is not None
        assert t.description == 'NewSTWithSalesTax'
        assert t.amount == Decimal('250.00')
        assert t.sales_tax == Decimal('12.50')
        assert t.day_of_month == 15

    def test_60_add_weekly_modal_on_click(self, base_url, selenium):
        """Test adding a new weekly scheduled transaction via modal"""
        self.baseurl = base_url
        self.get(selenium, base_url + '/scheduled')
        link = selenium.find_element(By.ID, 'btn_add_sched')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Scheduled Transaction'
        desc = body.find_element(By.ID, 'sched_frm_description')
        desc.send_keys('NewWeeklyTest')
        _type = body.find_element(By.ID, 'sched_frm_type_weekly')
        _type.click()
        day_of_week_sel = Select(body.find_element(By.ID, 'sched_frm_day_of_week'))
        assert day_of_week_sel.first_selected_option.text == 'Monday'
        day_of_week_sel.select_by_value('2')  # Wednesday
        amt = body.find_element(By.ID, 'sched_frm_amount')
        amt.send_keys('75.50')
        acct_sel = Select(body.find_element(By.ID, 'sched_frm_account'))
        acct_sel.select_by_value('1')
        budget_sel = Select(body.find_element(By.ID, 'sched_frm_budget'))
        budget_sel.select_by_value('1')
        notes = body.find_element(By.ID, 'sched_frm_notes')
        notes.send_keys('weekly test notes')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert 'Successfully saved ScheduledTransaction' in x.text.strip()
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # test that the new transaction appears in the table
        table = selenium.find_element(By.ID, 'table-scheduled-txn')
        texts = [y[4] for y in self.tbody2textlist(table)]
        assert 'NewWeeklyTest' in texts

    def test_61_add_weekly_verify_db(self, testdb):
        """Verify new weekly scheduled transaction in database"""
        t = testdb.query(ScheduledTransaction).filter(
            ScheduledTransaction.description == 'NewWeeklyTest'
        ).first()
        assert t is not None
        assert t.description == 'NewWeeklyTest'
        assert t.num_per_period is None
        assert t.date is None
        assert t.day_of_month is None
        assert t.day_of_week == 2  # Wednesday
        assert t.annual_month is None
        assert t.annual_day is None
        assert t.amount == Decimal('75.50')
        assert t.account_id == 1
        assert t.budget_id == 1
        assert t.notes == 'weekly test notes'
        assert t.is_active is True

    def test_70_add_annual_modal_on_click(self, base_url, selenium):
        """Test adding a new annual scheduled transaction via modal"""
        self.baseurl = base_url
        self.get(selenium, base_url + '/scheduled')
        link = selenium.find_element(By.ID, 'btn_add_sched')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Scheduled Transaction'
        desc = body.find_element(By.ID, 'sched_frm_description')
        desc.send_keys('NewAnnualTest')
        _type = body.find_element(By.ID, 'sched_frm_type_annual')
        _type.click()
        month_sel = Select(body.find_element(By.ID, 'sched_frm_annual_month'))
        month_sel.select_by_value('7')  # July
        day_input = body.find_element(By.ID, 'sched_frm_annual_day')
        day_input.send_keys('4')  # July 4th
        amt = body.find_element(By.ID, 'sched_frm_amount')
        amt.send_keys('500.00')
        acct_sel = Select(body.find_element(By.ID, 'sched_frm_account'))
        acct_sel.select_by_value('2')
        budget_sel = Select(body.find_element(By.ID, 'sched_frm_budget'))
        budget_sel.select_by_value('4')
        notes = body.find_element(By.ID, 'sched_frm_notes')
        notes.send_keys('annual test notes')
        # submit the form
        selenium.find_element(By.ID, 'modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements(By.TAG_NAME, 'div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert 'Successfully saved ScheduledTransaction' in x.text.strip()
        # dismiss the modal
        selenium.find_element(By.ID, 'modalCloseButton').click()
        self.wait_for_jquery_done(selenium)
        # Note: we don't check the table here because with 10+ items the
        # DataTable pagination may not show the new item. The database
        # verification in test_71 confirms the transaction was created.

    def test_71_add_annual_verify_db(self, testdb):
        """Verify new annual scheduled transaction in database"""
        t = testdb.query(ScheduledTransaction).filter(
            ScheduledTransaction.description == 'NewAnnualTest'
        ).first()
        assert t is not None
        assert t.description == 'NewAnnualTest'
        assert t.num_per_period is None
        assert t.date is None
        assert t.day_of_month is None
        assert t.day_of_week is None
        assert t.annual_month == 7  # July
        assert t.annual_day == 4
        assert t.amount == Decimal('500.00')
        assert t.account_id == 2
        assert t.budget_id == 4
        assert t.notes == 'annual test notes'
        assert t.is_active is True
